
# https://github.com/terraref/tutorials/blob/geostreams-guide/sensors/06-list-datasets-by-plot.md
import os
import requests

import logging
log = logging.getLogger(__name__)

from terrautils.geostreams import get_sensor_by_name


# TODO this should be from the pyclowder package
def get_sensor_list(connection, host, key):
    """ Return a list of sensors from the geostream database
    
    Sensors are best thought of as data products can include derived data
    in addition to data collected by sensors. Sensor names returned 
    may include a plot id number associated with a sitename.

    Keyword arguments:
    connection -- connection information, used to get missing 
      parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    """

    url = "%sapi/geostreams/streams?key=%s" % (host, key)
    r = requests.get(url)
    r.raise_for_status()
    return r.json()


def unique_sensor_names(sensors):
    """ Return a list of unique sensor names from a sensor list

    Takes a list of sensors (get_sensor_list) and creates a set
    of unique names by removing the plot id.
    """

    rsp = set()
    for s in sensors:
        if s['name'].endswith(')'):
            rsp.add(s['name'].split('(')[0].strip())
        else:
            rsp.add(s['name'])

    return list(rsp)


def get_datapoints(connection, host, key, sensor_id, **params):
    """ Return a list of datapoints

    Keyword arguments:
    connection -- connection information, used to get missing 
      parameters and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    sensor_id -- the 
    """

    params['key'] = key
    params['stream_id'] = sensor

    url = "%sapi/geostreams/datapoints" % host
    r = requests.get(url, params=params)
    r.raise_for_status()
    return r.json()


def get_sensor(connection, host, key, sensor, sitename=''):
    """ Return the geostream stream dictionary for the sensor.

    Matches the specific sensor name. If sitename is given an 
    additional query is made to determine the plot (sitename) id and
    it is automatically append to the sensor name.

    Keyword arguments:
    connection -- connection information, used to get missing parameters
      and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    sensor -- the name of the sensor
    sitename -- plot name from betydb (optional)
    """

    # if sitename is given, look up id and append to sensor name
    if sitename:
        s = get_sensor_by_name(None, host, key, sitename)
        if s:
            plotid = s['id']
            if not sensor.endswith(')'):
                sensor += ' ({})'.format(plotid)
        else:
            return None

    log.debug('full sensor name = %s', sensor)
    
    url = '%sapi/geostreams/streams' % host
    params = { 'key': key, 'stream_name': sensor }
    r = requests.get(url, params=params)
    r.raise_for_status()
    return r.json()

# TODO this is probably a pyclowder function
# NOTE this function extracts the dataset id from the URI and creates
# a new URI based on the host variable. Is this the best approach?
def get_files(connection, host, key, dataset):
    """ Returns a list of files for the given dataset.

    Keyword arguments:
    connection -- connection information, used to get missing parameters
      and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    dataset - a URI to the dataset on clowder
    """

    log.debug('original dataset uri = %s', dataset)

    # extract the id from the dataset URI
    dataset_id = dataset.split('/')[-1]
    if dataset_id == 'files':
        dataset_id = dataset.split('/')[-2]

    url = '%sapi/datasets/%s/files' % (host, dataset_id)
    log.debug('new url = %s', url)
    r = requests.get(url, params={'key': key})
    r.raise_for_status()
    return r.json()


def get_file_listing(connection, host, key, sensor, sitename, 
                     since='', until=''):
    """ Return a list of clowder file records for a sensor.
    
    Queries geostrteams to get a list of datasets associated with a
    sensor. If sitename is given it is automatically append to the sensor
    name. The since and until parameters can be used to limit the time
    range.

    After getting a list of datasets, a Clowder query is made for each
    dataset to get the list of files. The aggregate list is returned.

    Keyword arguments:
    connection -- connection information, used to get missing parameters 
      and send status updates
    host -- the clowder host, including http and port, should end with a /
    key -- the secret key to login to clowder
    sensor -- the name of the sensor
    sitename -- plot name from betydb
    since -- starting time (optional)
    until -- ending time (optional)
    """

    files = []

    sens = get_sensor(connection, host, key, sensor, sitename)
    if sens:
        if 'status' in sens and sens['status'] == 'No data found':
            log.info("No sensor found for %s" % sensor+" - "+sitename)
            return []


        stream_id = sens[0]['id']

        url = '%sapi/geostreams/datapoints' % host
        params = { 'key': key, 'stream_id': stream_id }
        if since:
            params['since'] = since
        if until:
            params['until'] = until

        r = requests.get(url, params=params)
        r.raise_for_status()

        if len(r.json()) > 0:
            datasets = [ds['properties']['source_dataset'] for ds in r.json()]

            for ds in datasets:
                flist = get_files(connection, host, key, ds)
                if flist:
                    files.extend(flist)
        else:
            log.info("No datasets found for %s" % sensor+" - "+sitename)
    else:
        log.info("No sensor found for %s" % sensor+" - "+sitename)

    return files


def extract_file_paths(listing):
    """Takes file listing and returns a list of absolute file paths"""

    return [os.path.join(l['filepath'], l['filename']) for l in listing]




from setuptools import setup, find_packages

def description():
    with open('readme.rst', encoding="utf-8") as f:
         return f.read()

setup(name='terrautils',
      packages=find_packages(),
      version='1.5.5',
      include_package_data=True,
      description='TERRA-REF workflow utilities',
      long_description=description(),
      author='Max Burnette',
      author_email='mburnet2@illinois.edu',

      url='https://terraref.org',
      project_urls = {
        'Source': 'https://github.com/terraref/terrautils',
        'Tracker': 'https://github.com/terraref/computing-pipeline/issues',
      },

      install_requires=[
          'cfunits',
          'influxdb',
          'matplotlib',
          'netCDF4',
          'numpy',
          'Pillow',
          'python-dateutil',
          'scipy',
          'utm',
          'python-logstash',
          'pyclowder>=2,<3'
      ],
      zip_safe=False,

      license='BSD',
      classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Database',
        'Topic :: Scientific/Engineering :: GIS',
        'Topic :: Utilities',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
      ],
      keywords=['terraref', 'clowder', 'field crop', 'phenomics', 'computer vision', 'remote sensing']
)




"""Metadata

This module provides useful reference methods for accessing and cleaning TERRA-REF metadata.
"""

import terrautils.lemnatec

from terrautils.secure import decrypt_pipeline_string

def clean_metadata(json, sensorId, fixed=False):
    """ Given a metadata object, returns a cleaned object with standardized structure 
        and names.
    """
    cleaned = clean_json_keys(json)
    if 'lemnatec_measurement_metadata' in json.keys():
        cleaned = terrautils.lemnatec.clean(cleaned, sensorId, fixed=fixed)
    else:
        return None

    if 'terraref_cleaned_metadata' not in cleaned:
        cleaned["terraref_cleaned_metadata"] = True
    return cleaned


def clean_json_keys(jsonobj):
    """If metadata keys have periods in them, Clowder will reject the metadata.
    """
    clean_json = {}
    for key in jsonobj.keys():
        try:
            jsonobj[key].keys() # Is this a json object?
            clean_json[key.replace(".","_")] = clean_json_keys(jsonobj[key])
        except:
            clean_json[key.replace(".","_")] = jsonobj[key]

    return clean_json


def calculate_scan_time(metadata):
    """Parse scan time from metadata.

        Returns:
            timestamp string
    """
    scan_time = None

    if 'terraref_cleaned_metadata' in metadata and metadata['terraref_cleaned_metadata']:
        scan_time = metadata['gantry_variable_metadata']['datetime']
    else:
        for sub_metadata in metadata:
            if 'content' in sub_metadata:
                sub_metadata = sub_metadata['content']
            if 'terraref_cleaned_metadata' in sub_metadata and sub_metadata['terraref_cleaned_metadata']:
                scan_time = sub_metadata['gantry_variable_metadata']['datetime']

    return scan_time


def get_terraref_metadata(clowder_md, sensor_id=None, station='ua-mac'):
    """Crawl Clowder metadata object and return TERRARef metadata or None.

    If sensor_id given, will attach fixed sensor metadata from that sensor."""

    terra_md = {}

    if 'terraref_cleaned_metadata' in clowder_md and clowder_md['terraref_cleaned_metadata']:
        terra_md = clowder_md
    else:
        for sub_metadata in clowder_md:
            if 'content' in sub_metadata:
                sub_metadata = sub_metadata['content']
            if 'terraref_cleaned_metadata' in sub_metadata and sub_metadata['terraref_cleaned_metadata']:
                terra_md = sub_metadata

    # Add sensor fixed metadata
    if sensor_id:
        query_date = get_date_from_cleaned_metadata(terra_md)
        sensor_fixed = get_sensor_fixed_metadata(sensor_id, query_date)
        if 'sensor_fixed_metadata' in terra_md:
            sensor_fixed['url'] = terra_md['sensor_fixed_metadata']['url']
        terra_md['sensor_fixed_metadata'] = sensor_fixed

    return terra_md


def get_extractor_metadata(clowder_md, extractor_name, extractor_version=None):
    """Crawl Clowder metadata object for particular extractor metadata and return if found.

    If extractor_version specified, returned metadata must match."""
    for sub_metadata in clowder_md:
        if 'agent' in sub_metadata:
            agent_data = sub_metadata['agent']
            if 'name' in agent_data and agent_data['name'].find(extractor_name) > -1:
                if not extractor_version:
                    return sub_metadata['content']
                else:
                    # TODO: Eventually check this in preferred way
                    if 'extractor_version' in sub_metadata['content']:
                        existing_ver = str(sub_metadata['content']['extractor_version'])
                        if existing_ver == extractor_version:
                            return sub_metadata['content']

    return None


# pylint: disable=too-many-branches
def pipeline_get_metadata(metadata):
    """Look through the metadata looking for pipeline configuration
    Args:
        metadata(JSON): the JSON object, or list of JSON objects, to search
    Returns:
        The found JSON is returned, otherwise None is returned. None is also returned
        if the passed in JSON is invalid or is empty.
    Notes:
        If the metadata parameter is an a list, it will be iterated over and each list element
        inspected for pipeline data. The first found instance is returned.
        If the metadata parameter is not a list, it's inspected to see if it contains
        pipeline data, and that's returned if found.

        There are two keys looked for off the JSON passed in, as follows, in order:
            1) "content" -> "pipeline"
            2) "pipeline"
    """
    found_metadata = None

    # Check the parameter
    if not metadata:
        return found_metadata

    json_len = len(metadata)
    if json_len <= 0:
        return found_metadata

    try:
        # Look for a list of JSON
        if isinstance(metadata, list):
            for one_metadata in metadata:
                if 'content' in one_metadata:
                    if 'pipeline' in one_metadata['content']:
                        found_metadata = one_metadata['content']['pipeline']
                        break

            if found_metadata is None:
                for one_metadata in metadata:
                    if 'pipeline' in one_metadata:
                        found_metadata = one_metadata['pipeline']
                        break

        elif 'content' in metadata:
            if 'pipeline' in metadata['content']:
                found_metadata = one_metadata['content']['pipeline']

        elif 'pipeline' in metadata:
            found_metadata = metadata['pipeline']
    finally:
        pass

    # Check if we need to perform additional actions on the metadata
    if found_metadata and "clowder" in found_metadata:
        if "password" in found_metadata["clowder"]:
            clowder_pass = found_metadata["clowder"]["password"]
            if clowder_pass.startswith("secured:"):
                plain_pass = decrypt_pipeline_string(clowder_pass[8:])
                if not plain_pass is None:
                    found_metadata["clowder"]["password"] = plain_pass

    return found_metadata

def prepare_pipeline_metadata(metadata):
    """Fixes the metadata so that the drone pipeline easily reference it
    Args:
        metadata(JSON): the JSON object to format
    Returns:
        A deep copy of the metadata with any necessary changes made.
    """
    from copy import deepcopy
    from terrautils.secure import encrypt_pipeline_string

    return_metadata = deepcopy(metadata)
    if "clowder" in return_metadata:
        if "password" in return_metadata["clowder"]:
            encrypted = encrypt_pipeline_string(return_metadata["clowder"]["password"])
            if not encrypted is None:
                return_metadata["clowder"]["password"] = "secured:" + encrypted
            else:
                return_metadata["clowder"]["password"] = "<removed>"
            
    return {"pipeline" : return_metadata}


def get_season_and_experiment(timestamp, sensor, terra_md_full):
    """Attempts to extract season & experiment from TERRA-REF metadata given timestamp.

    If the values weren't in TERRA metadata but were fetched from BETY, updated experiment will be
    returned as well.
    """
    season_name, experiment_name, expmd = "Unknown Season", "Unknown Experiment", None
    if 'experiment_metadata' in terra_md_full and len(terra_md_full['experiment_metadata']) > 0:
        for experiment in terra_md_full['experiment_metadata']:
            if 'name' in experiment:
                if ":" in experiment['name']:
                    season_name = experiment['name'].split(": ")[0]
                    experiment_name = experiment['name'].split(": ")[1]
                else:
                    experiment_name = experiment['name']
                    season_name = None
                break
    else:
        # Try to determine experiment data dynamically
        expmd = terrautils.lemnatec._get_experiment_metadata(timestamp.split("__")[0], sensor)
        if len(expmd) > 0:
            for experiment in expmd:
                if 'name' in experiment:
                    if ":" in experiment['name']:
                        season_name = experiment['name'].split(": ")[0]
                        experiment_name = experiment['name'].split(": ")[1]
                    else:
                        experiment_name = experiment['name']
                        season_name = None
                    break

    return (season_name, experiment_name, expmd)


def get_preferred_synonym(variable):
    """Execute a thesaurus check to see if input variable has alternate preferred name."""
    pass


def get_sensor_fixed_metadata(sensor_id, query_date):
    """Get fixed sensor metadata from Clowder."""
    return terrautils.lemnatec._get_sensor_fixed_metadata(sensor_id, query_date)


def get_date_from_cleaned_metadata(md):
    default = "2012-01-01"
    if "gantry_variable_metadata" in md:
        if "date" in md["gantry_variable_metadata"]:
            return md["gantry_variable_metadata"]["date"]
        else:
            return default
    else:
        return default



"""BETYdb

This module provides wrappers to BETY API for getting and posting data.
"""

import os
import logging
from datetime import datetime

import json
import requests
from osgeo import ogr

from terrautils.spatial import geometry_to_geojson

BETYDB_URL = "https://terraref.ncsa.illinois.edu/bety"
BETYDB_LOCAL_CACHE_FOLDER = os.environ.get('BETYDB_LOCAL_CACHE_FOLDER', '/home/extractor/')

BETYDB_CULTIVARS = None
BETYDB_TRAITS = None
BETYDB_EXPERIMENTS = None


def add_arguments(parser):
    """Adds BETYdb related arguments to the command line argument parser

    parser - the argument parser to add our definitions to
    """
    parser.add_argument('--betyURL', dest="bety_url", type=str, nargs='?',
                        default="https://terraref.ncsa.illinois.edu/bety/api/v1/traits.csv",
                        help="traits API endpoint of BETY instance that outputs should be posted to")

    parser.add_argument('--betyKey', dest="bety_key", type=str, nargs='?',
                        default=os.getenv('BETYDB_KEY', ''),
                        help="API key for BETY instance specified by betyURL")


def get_bety_key():
    """return key from environment or ~/.betykey if it exists."""

    key = os.environ.get('BETYDB_KEY', '')
    if key:
        return key

    keyfile_path = os.path.expanduser('~/.betykey')
    if os.path.exists(keyfile_path):
        keyfile = open(keyfile_path, "r")
        return keyfile.readline().strip()

    else:
        raise RuntimeError("BETYDB_KEY not found. Set environmental variable " +
                           "or create $HOME/.betykey.")


def get_bety_url(path=''):
    """return betydb url from environment with optional path

    Of 3 options string join, os.path.join and urlparse.urljoin, os.path.join
    is the best at handling excessive / characters.
    """

    url = os.environ.get('BETYDB_URL', BETYDB_URL)
    return os.path.join(url, path)


def get_bety_api(endpoint=None):
    """return betydb API based on betydb url"""

    url = get_bety_url(path='api/v1/{}'.format(endpoint))
    return url


def query(endpoint="search", **kwargs):
    """return betydb API results.

    This is general function for querying the betyDB API. It automatically
    decodes the json response if one is returned.
    """

    payload = {'key': get_bety_key()}
    payload.update(kwargs)

    req = requests.get(get_bety_api(endpoint), params=payload)
    req.raise_for_status()
    return req.json()

def search(**kwargs):
    """Return cleaned up array from query() for the search table."""

    query_data = query(**kwargs)
    if query_data:
        return [view["traits_and_yields_view"] for view in query_data['data']]
    return []

def get_cultivars(**kwargs):
    """Return cleaned up array from query() for the cultivars table.
        If global variable isn't populated, check if a local file is present and read from it if so.
        This is for deployments where data is pre-fetched (e.g. for a Condor job).
        Otherwise the BETY API will be called.
        In either case, data will be kept in memory for subsequent calls.
    """
    global BETYDB_CULTIVARS

    if BETYDB_CULTIVARS is None:
        cache_file = os.path.join(BETYDB_LOCAL_CACHE_FOLDER, "bety_cultivars.json")
        if os.path.exists(cache_file):
            with open(cache_file) as infile:
                query_data = json.load(infile)
                if query_data:
                    BETYDB_CULTIVARS = query_data
                    return [t["cultivar"] for t in query_data['data']]
        else:
            query_data = query(endpoint="cultivars", **kwargs)
            if query_data:
                BETYDB_CULTIVARS = query_data
                return [t["cultivar"] for t in query_data['data']]
    else:
        return [t["cultivar"] for t in BETYDB_CULTIVARS['data']]

    return []

def dump_cultivars(**kwargs):
    """Generate bety_cultivars.json file"""
    query_data = query(endpoint="cultivars", limit='none', **kwargs)
    if query_data:
        cache_file = os.path.join(BETYDB_LOCAL_CACHE_FOLDER, "bety_cultivars.json")
        with open(cache_file, 'w') as cache_out:
            cache_out.write(json.dumps(query_data))


def get_experiments(**kwargs):
    """Return cleaned up array from query() for the experiments table.
        If global variable isn't populated, check if a local file is present and read from it if so.
        This is for deployments where data is pre-fetched (e.g. for a Condor job).
        Otherwise the BETY API will be called.
        In either case, data will be kept in memory for subsequent calls.
    """
    global BETYDB_EXPERIMENTS

    if BETYDB_EXPERIMENTS is None:
        cache_file = os.path.join(BETYDB_LOCAL_CACHE_FOLDER, "bety_experiments.json")
        if os.path.exists(cache_file):
            with open(cache_file) as infile:
                query_data = json.load(infile)
                if query_data:
                    if 'associations_mode' in kwargs:
                        BETYDB_EXPERIMENTS = query_data
                    return [t["experiment"] for t in query_data['data']]
        else:
            query_data = query(endpoint="experiments", **kwargs)
            if query_data:
                if 'associations_mode' in kwargs:
                    BETYDB_EXPERIMENTS = query_data
                return [t["experiment"] for t in query_data['data']]
    else:
        return [t["experiment"] for t in BETYDB_EXPERIMENTS['data']]

    return []

def dump_experiments(**kwargs):
    """Generate bety_experiments.json file"""
    query_data = query(endpoint="experiments", associations_mode='full_info', limit='none', **kwargs)
    if query_data:
        cache_file = os.path.join(BETYDB_LOCAL_CACHE_FOLDER, "bety_experiments.json")
        with open(cache_file, 'w') as cache_out:
            cache_out.write(json.dumps(query_data))


def get_trait(trait_id):
    """Returns python dictionary for a single trait."""
    query_data = get_traits(id=trait_id)
    if query_data:
        return query_data[0]

    return []


def get_traits(**kwargs):
    """Return cleaned up array from query() for the traits table.
        If global variable isn't populated, check if a local file is present and read from it if so.
        This is for deployments where data is pre-fetched (e.g. for a Condor job).
        Otherwise the BETY API will be called.
        In either case, data will be kept in memory for subsequent calls.
    """
    global BETYDB_TRAITS

    if BETYDB_TRAITS is None:
        cache_file = os.path.join(BETYDB_LOCAL_CACHE_FOLDER, "bety_traits.json")
        if os.path.exists(cache_file):
            with open(cache_file) as infile:
                query_data = json.load(infile)
                if query_data:
                    BETYDB_TRAITS = query_data
                    return [t["trait"] for t in query_data['data']]
        else:
            query_data = query(endpoint="traits", **kwargs)
            if query_data:
                BETYDB_TRAITS = query_data
                return [t["trait"] for t in query_data['data']]
    else:
        return [t["trait"] for t in BETYDB_TRAITS['data']]

    return []

def dump_traits(**kwargs):
    """Generate bety_traits.json file"""
    query_data = query(endpoint="traits", limit='none', **kwargs)
    if query_data:
        cache_file = os.path.join(BETYDB_LOCAL_CACHE_FOLDER, "bety_traits.json")
        with open(cache_file, 'w') as cache_out:
            cache_out.write(json.dumps(query_data))


def get_site(site_id):
    """Returns python dictionary for a single site"""
    query_data = get_sites(id=site_id)
    if query_data:
        return query_data[0]
    return []

def get_sites(filter_date='', include_halves=False, **kwargs):
    """Return a site array from query() from the sites table.

    e.g.
            get_sites(city="Maricopa")
            get_sites(sitename="MAC Field Scanner Season 4 Range 4 Column 6")
            get_sites(contains="-111.97496613200647,33.074671230742446")

      filter_date -- YYYY-MM-DD to filter sites to specific experiment by date
    """

    if not filter_date:
        # SCENARIO I - NO FILTER DATE
        # Basic query, efficient even with 'containing' parameter.
        query_data = query(endpoint="sites", limit='none', **kwargs)
        if query_data:
            return [t["site"] for t in query_data['data']]
    else:
        # SCENARIO II - YES FILTER DATE
        # Get experiments by date and return all associated sites, optionally filtering by location.
        targ_date = datetime.strptime(filter_date, '%Y-%m-%d')
        query_data = get_experiments(associations_mode='full_info', limit='none', **kwargs)
        if query_data:
            results = []
            for exp in query_data:
                start = datetime.strptime(exp['start_date'], '%Y-%m-%d')
                end = datetime.strptime(exp['end_date'], '%Y-%m-%d')
                if start <= targ_date <= end and 'sites' in exp:
                    for one_entry in exp['sites']:
                        site = one_entry['site']
                        # TODO: Eventually find better solution for S4 half-plots - they are omitted here
                        if (site["sitename"].endswith(" W") or site["sitename"].endswith(" E")) \
                                                                                    and not include_halves:
                            continue
                        if 'containing' in kwargs:
                            # Need to filter additionally by geometry
                            site_geom = ogr.CreateGeometryFromWkt(site['geometry'])
                            coords = kwargs['containing'].split(",")
                            pt_geom = ogr.CreateGeometryFromWkt("POINT(%s %s)" % (coords[1], coords[0]))
                            if site_geom.Intersects(pt_geom):
                                if site not in results:
                                    results.append(site)
                        else:
                            # If no containing parameter, include all sites
                            if site not in results:
                                results.append(site)
            return results
        else:
            logging.error("No experiment data could be retrieved.")

    return []

def get_sites_by_latlon(latlon, filter_date='', **kwargs):
    """Gets list of sites from BETYdb, filtered by a contained point.

      latlon (tuple) -- only sites that contain this point will be returned
      filter_date -- YYYY-MM-DD to filter sites to specific experiment by date
    """

    latlon_api_arg = "%s,%s" % (latlon[0], latlon[1])

    return get_sites(filter_date=filter_date, containing=latlon_api_arg, **kwargs)


def get_site_boundaries(filter_date='', **kwargs):
    """Get a dictionary of site GeoJSON bounding boxes filtered by standard arguments.

    filter_date -- YYYY-MM-DD to filter sites to specific experiment by date

    Returns:
        {
            'sitename_1': 'geojson bbox',
            'sitename_2': 'geojson bbox',
            ...
         }
    """

    sitelist = get_sites(filter_date, **kwargs)
    bboxes = {}

    for site in sitelist:
        geom = ogr.CreateGeometryFromWkt(site['geometry'])

        if geom:
            bboxes[site['sitename']] = geometry_to_geojson(geom, 'EPSG', '4326')
        else:
            logging.error("Site boundary geometry is invalid for site: %s", site['sitename'])

    return bboxes


def submit_traits(csv, filetype='csv', betykey='', betyurl=''):
    """ Submit traits file to BETY; can be CSV, JSON or XML."""

    # set defaults if necessary
    if not betykey:
        betykey = get_bety_key()
    if not betyurl:
        betyurl = get_bety_api('traits')

    request_payload = {'key':betykey}

    if filetype == 'csv':
        content_type = 'text/csv'
    elif filetype == 'json':
        content_type = 'application/json'
    elif filetype == 'xml':
        content_type = 'application/xml'
    else:
        logging.error("Unsupported file type.")
        return None

    resp = requests.post("%s.%s" % (betyurl, filetype), params=request_payload,
                         data=file(csv, 'rb').read(),
                         headers={'Content-type': content_type})

    if resp.status_code in [200, 201]:
        logging.info("Data successfully submitted to BETYdb.")
        return resp.json()['data']['ids_of_new_traits']
    else:
        logging.error("Error submitting data to BETYdb: %s -- %s", resp.status_code, resp.reason)
        resp.raise_for_status()

    return None



"""Contains routines specific to image files
"""
import logging
import subprocess

from osgeo import gdal, ogr
from numpy import nan

import osr

from terrautils.extractors import file_exists, load_json_file

# Returns if the MIME type of 'image' is found in the text passed in. A reasonable effort is
# made to identify the section containing the type by looking for the phrase 'Mime', or 'mime',
# or 'MIME' and using that as the basis for determining the type
def find_image_mime_type(text):
    """Looks for a MIME image type in the text passed in.

       It's expected that the MIME label may be something other than the string 'Mime type' so
       a proximity search is made for the MIME type label followed by 'image/'. If found, True
       is returned and False otherwise

    Args:
        text(str): The text in which to find a MIME type of 'image'

    Returns:
        None is returned if the string is empty or the MIME keyword is not found.
        False is returned if a MIME type of 'image' isn't found
        True is returned upon success
    """
    if not text:
        return None

    # Try to find the beginning and end of the mime type (but not the subtype)
    pos = text.find('Mime')
    if pos < 0:
        pos = text.find('mime')
    if pos < 0:
        pos = text.find('MIME')
    if pos < 0:
        return None

    end_pos = text.find('/', pos)
    if end_pos < 0:
        return False

    # Get the portion of the string containing the possible mime type making sure we have
    # something reasonable
    mime = text[pos : end_pos]
    mime_len = len(mime)
    if (mime_len > 50) or (mime.find('\n') >= 0) or (mime.find('\r') >= 0):
        return False

    # Look for a 'reasonable' image mime type
    if mime.endswith('image'):
        return True

    return False

# Determines if the file is an image type
def file_is_image_type(identify_binary, filename, metadata_filename=None):
    """Uses the identify application to generate the MIME type of the file and
       looks for an image MIME type. If a metadata filename is specified, the
       JSON in the file is loaded first and the MIME type is looked for. If
       the metadata filename is not specified, or a MIME type was not found in
       the metadata, the identity application is used.

    Args:
        identify_binary(str): path to the executable which will return a MIME type on an image file
        filename(str): the path to the file to check
        metadata_filename(str): the path to JSON metadata associated with the file in which to look
        for a 'contentType' tag containing the MIME type

    Returns:
        True is returned if the file is a MIME image type
        False is returned upon failure or the file is not a type of image
    """
    logger = logging.getLogger(__name__)

    # Try to determine the file type from its JSON information (metadata if from Clowder API)
    try:
        if metadata_filename and file_exists(metadata_filename):
            file_md = load_json_file(metadata_filename)
            if file_md:
                if 'contentType' in file_md:
                    if file_md['contentType'].startswith('image'):
                        return True
    # pylint: disable=broad-except
    except Exception as ex:
        logger.info("Exception caught: %s", str(ex))
    # pylint: enable=broad-except

    # Try to determine the file type locally
    try:
        is_image_type = find_image_mime_type(
            subprocess.check_output(
                [identify_binary, "-verbose", filename], stderr=subprocess.STDOUT))

        if not is_image_type is None:
            return is_image_type
    # pylint: disable=broad-except
    except Exception as ex:
        logger.info("Exception caught: %s", str(ex))
    # pylint: enable=broad-except

    return False

# Checks if the file has geometry associated with it and returns the bounds
def image_get_geobounds(filename):
    """Uses gdal functionality to retrieve recilinear boundaries from the file

    Args:
        filename(str): path of the file to get the boundaries from

    Returns:
        The upper-left and calculated lower-right boundaries of the image in a list upon success.
        The values are returned in following order: min_y, max_y, min_x, max_x. A list of numpy.nan
        is returned if the boundaries can't be determined
    """
    logger = logging.getLogger(__name__)

    try:
        # TODO: handle non-ortho images
        src = gdal.Open(filename)
        ulx, xres, _, uly, _, yres = src.GetGeoTransform()
        lrx = ulx + (src.RasterXSize * xres)
        lry = uly + (src.RasterYSize * yres)

        min_y = min(uly, lry)
        max_y = max(uly, lry)
        min_x = min(ulx, lrx)
        max_x = max(ulx, lrx)

        return [min_y, max_y, min_x, max_x]
    # pylint: disable=broad-except
    except Exception as ex:
        logger.info("[image_get_geobounds] Exception caught: %s", str(ex))
    # pylint: enable=broad-except

    return [nan, nan, nan, nan]

# Get the tuple from the passed in polygon
def polygon_to_tuples(polygon):
    """Convert polygon passed in to the following list:
        ( lat (y) min, lat (y) max,
          long (x) min, long (x) max)

    Args:
        polygon(object) - OGR Polygon (type ogr.wkbPolygon)

    Return:
        A tuple of (min Y, max Y, min X, max X)
    """
    logger = logging.getLogger(__name__)

    min_x, min_y, max_x, max_y = None, None, None, None
    try:
        if polygon.GetGeometryType() == ogr.wkbPolygon:
            ring = polygon.GetGeometryRef(0)
            point_count = ring.GetPointCount()
            for point_idx in range(point_count):
                pt_x, pt_y, _ = ring.GetPoint(point_idx)
                if min_x is None or pt_x < min_x:
                    min_x = pt_x
                if max_x is None or pt_x > max_x:
                    max_x = pt_x
                if min_y is None or pt_y < min_y:
                    min_y = pt_y
                if max_y is None or pt_y > max_y:
                    max_y = pt_y
    # pylint: disable=broad-except
    except Exception as ex:
        logger.warn("[polygon_to_tuples] Exception caught: %s", str(ex))
    # pylint: enable=broad-except

    return (min_y, max_y, min_x, max_x)

def polygon_to_tuples_transform(polygon, dest_spatial):
    """Transforms the polygon to the specified transformation if the polygon's coordinate system
       doesn't match. If there isn't a coordinate system associated with the polygon then no
       transformation is done.
    Args:
        polygon(object) - OGR Polygon (type ogr.wkbPolygon)
        dest_spatial(ogr.spatialReference): the spatial reference of target polygon values
    Return:
        A tuple of (min Y, max Y, min X, max X)
    Notes:
        The original polygon is unchanged whether or not a transformation takes place
    """
    logger = logging.getLogger(__name__)

    try:
        src_ref = polygon.GetSpatialReference()
        if src_ref and dest_spatial and not src_ref.IsSame(dest_spatial):
            transform = osr.CoordinateTransformation(src_ref, dest_spatial)
            new_src = polygon.Clone()
            if new_src:
                new_src.Transform(transform)
                return polygon_to_tuples(new_src)
            logger.error("[polygon_to_tuples_transform] Unable to create a polygon copy for " +
                         "coordinate transformation")
    # pylint: disable=broad-except
    except Exception as ex:
        logger.warn("[polygon_to_tuples_transform] Exception caught: %s", str(ex))
        logger.warn("[polygon_to_tuples_transform] returning non-transformed polygon points")
    # pylint: enable=broad-except

    return polygon_to_tuples(polygon)

def get_epsg(filename):
    """Returns the EPSG of the georeferenced image file
    Args:
        filename(str): path of the file to retrieve the EPSG code from
    Return:
        Returns the found EPSG code, or None if it's not found or an error ocurred
    """
    logger = logging.getLogger(__name__)

    try:
        src = gdal.Open(filename)

        proj = osr.SpatialReference(wkt=src.GetProjection())

        return proj.GetAttrValue('AUTHORITY', 1)
    # pylint: disable=broad-except
    except Exception as ex:
        logger.warn("[get_epsg] Exception caught: %s", str(ex))
    # pylint: enable=broad-except

    return None



"""Extractors

This module contains dictionaries and references for handling sensor information.
"""

import os
import re
import datetime

from terrautils.betydb import get_experiments

# 2017
year_p = '(20\d\d)'
# 06
month_p = '(0[1-9]|1[0-2])'
# 28
day_p = '(10|20|[0-2][1-9]|3[01])'
# 2017-06-28
date_p = '{}-{}-{}'.format(year_p, month_p, day_p)
# 23-48-28
time_p = '([0-1]\d|2[0-3])-([0-5]\d)-([0-5]\d)'
# 23-48-28-435
full_time_p = '([0-1]\d|2[0-3])-([0-5]\d)-([0-5]\d)-(\d{3})'
# 2017-06-28__23-48-28-435
full_date_p = '{}__{}'.format(date_p, full_time_p)

""" If the bety traits that are submitted to bety need to be renamed, only
    change the value of the dict key/value pair. Extractors will reference
    the key in their code and use whatever the value is for trait name."""
STATIONS = {
    'danforth': {
        'ddpscIndoorSuite': {
            "bety_traits": {
                "sv_area": "sv_area",
                "tv_area": "tv_area",
                "hull_area": "hull_area",
                "solidity": "solidity",
                "height": "height",
                "permieter": "perimeter"
            }
        }
    },

    'ksu': {
        'dsm': {
            'display': 'Digital Surface Model GeoTIFFs',
            'template': '{base}/{station}/Level_1/{sensor}/{filename}',
            'pattern': '{time}_DSM_16ASH-TERRA.tif'
        },
        'rededge': {
            'display': 'Red Edge',
            'template': '{base}/{station}/Level_1/{sensor}/{filename}',
            'pattern': '{time}_BGREN_16ASH-TERRA'
        }
    },

    'ua-mac': {
        'co2Sensor': {
            "fixed_metadata_json": "/sensors/co2sensor/sensor_fixed_metadata.json",
            "fixed_metadata_datasetid": "5873a9924f0cad7d8131b648",
            'template': '{base}/{station}/raw_data/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': '([0-9a-f]){8}-([0-9a-f]){4}-([0-9' + \
                       'a-f]){4}-([0-9a-f]){4}-([0-9a-f])' + \
                       '{12}_(metadata.json|rawData0000.bin)'
        },
        
        'stereoTop': {
            "fixed_metadata_json": "/sensors/stereo/sensor_fixed_metadata.json",
            "fixed_metadata_datasetid": "5873a8ae4f0cad7d8131ac0e",
            'template': '{base}/{station}/raw_data/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': 'hashid{opts}.bin'

        },
        
        'flirIrCamera': {
            "fixed_metadata_json": "/sensors/flirIrCamera/sensor_fixed_metadata.json",
            "fixed_metadata_datasetid": "5873a7184f0cad7d8131994a",
            'template': '{base}/{station}/raw_data/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': 'hashid_ir.bin'
        },
        
        "cropCircle": {
            "fixed_metadata_json": "/sensors/cropCircle/sensor_fixed_metadata.json",
            "fixed_metadata_datasetid": "5873a7ed4f0cad7d8131a2e7"
        },
        
        "irrigation": {
            "fixed_metadata_json": "/sensors/weatherStation/sensor_fixed_metadata.json",
            "fixed_metadata_datasetid": "TBD",
            "geostream": ""
        },
        
        'EnvironmentLogger': {
            'template': '{base}/{station}/raw_data/' + \
                        '{sensor}/{date}/{filename}',
            'pattern': '{timestamp}_environment_logger.json',
            'fixed_metadata_datasetid': 'TBD'
        },
        
        "lightning": {
            "fixed_metadata_json": "/sensors/weatherStation/sensor_fixed_metadata.json",
            "fixed_metadata_datasetid": "TBD",
            "geostream": ""
        },
        
        "ndviSensor": {
            "fixed_metadata_json": "/sensors/ndviSensor/sensor_fixed_metadata.json",
            "fixed_metadata_datasetid": "5873a8f64f0cad7d8131af54"
        },
        
        "parSensor": {
            "fixed_metadata_json": "/sensors/parSensor/sensor_fixed_metadata.json",
            "fixed_metadata_datasetid": "5873a8ce4f0cad7d8131ad86"
        },
        
        "priSensor": {
            "fixed_metadata_json": "/sensors/priSensor/sensor_fixed_metadata.json",
            "fixed_metadata_datasetid": "5873a9174f0cad7d8131b09a"
        },
        
        "ps2Top": {
            "fixed_metadata_json": "/sensors/ps2/sensor_fixed_metadata.json",
            "fixed_metadata_datasetid": "5873a84b4f0cad7d8131a73d",
            'template': '{base}/{station}/raw_data/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': 'hashid_rawData{opts}.bin'
        },
        
        "scanner3DTop": {
            "fixed_metadata_json": "/sensors/scanner3D/sensor_fixed_metadata.json",
            "fixed_metadata_datasetid": "5873a7444f0cad7d81319b2b",
            'template': '{base}/{station}/Level_1/' + \
                        'laser3d_las/{date}/{timestamp}/{filename}',
            'pattern': 'hashid__Top-heading-{opts}_0.ply'
        },
        
        "SWIR": {
            "fixed_metadata_json": "/sensors/SWIR/sensor_fixed_metadata.json",
            "fixed_metadata_datasetid": "5873a79e4f0cad7d81319f5f"
        },
        
        "VNIR": {
            "fixed_metadata_json": "/sensors/VNIR/sensor_fixed_metadata.json",
            "fixed_metadata_datasetid": "5873a7bb4f0cad7d8131a0b7"
        },
        
        "weather": {
            "fixed_metadata_json": "/sensors/weatherStation/sensor_fixed_metadata.json",
            "fixed_metadata_datasetid": "TBD"
        },
      
        "scanalyzer": {
            "fixed_metadata_json": "/sensors/scanalyzer/sensor_fixed_metadata.json",
            "fixed_metadata_datasetid": "5873eac64f0cad7d81349b0b"
        },

        'rgb_fullfield': {
            'display': 'Full Field Stitched Mosaics',
            'template': '{base}/{station}/Level_2/' + \
                        '{sensor}/{date}/{filename}',
            'pattern': '{sensor}_L2_{station}_{date}{opts}.tif'
        },

        'ir_fullfield': {
            'display': 'Full Field Stitched Mosaics',
            'template': '{base}/{station}/Level_2/' + \
                        '{sensor}/{date}/{filename}',
            'pattern': '{sensor}_L2_{station}_{date}{opts}.tif'
        },

        'laser3d_fullfield': {
            'display': 'Full Field Stitched Mosaics',
            'template': '{base}/{station}/Level_2/' + \
                        '{sensor}/{date}/{filename}',
            'pattern': '{sensor}_L2_{station}_{date}{opts}.tif'
        },

        'vnir_netcdf': {
            'display': 'VNIR Hyperspectral NetCDFs',
            'template': '{base}/{station}/Level_1/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': '{sensor}_L1_{station}_{timestamp}{opts}.nc',
            "bety_traits": {
                "NDVI705": "NDVI705"
            }
        },

        'swir_netcdf': {
            'display': 'SWIR Hyperspectral NetCDFs',
            'template': '{base}/{station}/Level_1/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': '{sensor}_L1_{station}_{timestamp}{opts}.nc'
        },

        'vnir_traits': {
            'template': '{base}/{station}/Level_3/' + \
                        '{sensor}/{date}/{filename}',
            'pattern': '{sensor}_L3_{station}_{date}{opts}.csv',
            'bety_traits': {
                'NDVI705': 'NDVI705'
            }
        },

        'swir_traits': {
            'template': '{base}/{station}/Level_3/' + \
                        '{sensor}/{date}/{filename}',
            'pattern': '{sensor}_L3_{station}_{date}{opts}.csv',
            'bety_traits': {
                'NDVI705': 'NDVI705'
            }
        },

        'rgb_geotiff': {
            'display': 'RGB GeoTIFFs',
            'template': '{base}/{station}/Level_1/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': '{sensor}_L1_{station}_{timestamp}{opts}.tif',
        },

        'rgb_nrmac': {
            'display': 'RGB GeoTIFFs (NRMAC Quality Score)',
            'template': '{base}/{station}/Level_2/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': '{sensor}_L2_{station}_{timestamp}{opts}_nrmac.tif',
        },

        'rgb_enhanced': {
            'display': 'RGB GeoTIFFs (Enhanced)',
            'template': '{base}/{station}/Level_2/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': '{sensor}_L2_{station}_{timestamp}{opts}_enhanced.tif',
        },

        'rgb_mask': {
            'display': 'RGB GeoTIFFs (Masked)',
            'template': '{base}/{station}/Level_2/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': '{sensor}_L2_{station}_{timestamp}{opts}_mask.tif',
        },

        'rgb_canopycover': {
            'template': '{base}/{station}/Level_3/' + \
                        '{sensor}/{date}/{filename}',
            'pattern': '{sensor}_L3_{station}_{date}{opts}.csv',
            'bety_traits': {
                'canopy_cover': 'canopy_cover'
            },
            "url": ""
        },

        'rgb_textureanalysis': {
            'display': 'RGB Texture Analysis',
            'template': '{base}/{station}/Level_3/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': '{sensor}_L3_{station}_{timestamp}{opts}.csv',
        },

        'ir_geotiff': {
            'display': 'Thermal IR GeoTIFFs',
            'template': '{base}/{station}/Level_1/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': '{sensor}_L1_{station}_{timestamp}{opts}.tif',
        },

        'ps2_png': {
            'display': 'PSII PNGs',
            'template': '{base}/{station}/Level_1/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': '{sensor}_L1_{station}_{timestamp}{opts}.png',
        },

        'ps2_fluorescence': {
            'display': 'PSII Fluorescence Features',
            'template': '{base}/{station}/Level_2/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': '{sensor}_L2_{station}_{timestamp}{opts}.png',
        },

        'spectral_index_csvs': {
            'display': 'Multispectral Index CSVs',
            'template': '{base}/{station}/Level_1/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': '{sensor}_L1_{station}_{timestamp}{opts}.csv',
        },

        'envlog_netcdf': {
            'display': 'EnvironmentLogger netCDFs',
            'template': '{base}/{station}/Level_1/' + \
                        '{sensor}/{date}/{filename}',
            'pattern': '{sensor}_L1_{station}_{timestamp}{opts}.nc',
        },

        'vnir_soil_masks': {
            'display': 'VNIR Soil Masks',
            'template': '{base}/{station}/Level_2/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': 'VNIR_L2_{station}_{timestamp}{opts}.nc'
        },

        'swir_soil_masks': {
            'display': 'SWIR Soil Masks',
            'template': '{base}/{station}/Level_2/'
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': 'SWIR_L2_{station}_{timestamp}{opts}.nc'
        },

        'laser3d_las': {
            'display': 'Laser Scanner 3D LAS',
            'template': '{base}/{station}/Level_1/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': 'scanner3DTop_L1_{station}_{timestamp}' + \
                       '_merged{opts}.las'
        },

        'laser3d_canopyheight': {
            'display': 'Laser Scanner 3D Canopy Height',
            'template': '{base}/{station}/Level_3/' + \
                        '{sensor}/{date}/{plot}/{filename}'
        },

        'laser3d_heightmap': {
            'display': 'Digital Surface Model GeoTiffs',
            'template': '{base}/{station}/Level_2/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': 'scanner3DTop_L2_{station}_{timestamp}' + \
                       '_heightmap{opts}.tif'
        },

        'weather_datparser': {
            "display": "UA-MAC AZMET Weather Station",
            "url": ""
        },

        'irrigation_datparser': {
            "display": "UA-MAC AZMET Weather Station",
            "url": ""
        },

        'energyfarm_datparser': {
            "display": "UIUC Energy Farm",
            "url": ""
        },

        'netcdf_metadata': {
            'template': '{base}/{station}/Level_1/' + \
                        '{sensor}/{date}/{timestamp}/{filename}',
            'pattern': '{sensor}_L2_{station}_{timestamp}{opts}.tif'
        },

        'ir_meantemp': {
            'template': '{base}/{station}/Level_3/' + \
                        '{sensor}/{date}/{filename}',
            'pattern': '{sensor}_L3_{station}_{date}{opts}.csv',
            'bety_traits': {
                'surface_temperature': 'surface_temperature'
            },
        },

        'plotclipper': {
            'display': 'Plot Clipper',
            'template': '{base}/{station}/Level_1_Plots/' + \
                        '{subsensor}/{date}/{plot}/{filename}'
        }
    }
}


def add_arguments(parser):

    if os.path.exists('/projects/arpae/terraref/sites'):
        TERRAREF_BASE = '/projects/arpae/terraref/sites'
    elif os.path.exists('/home/clowder/sites'):
        TERRAREF_BASE = '/home/clowder/sites'
    else:
        TERRAREF_BASE = '/home/extractor/sites'

    parser.add_argument('--terraref_base', type=str,
            default=os.environ.get('TERRAREF_BASE', TERRAREF_BASE),
            help='path to the sites directory on local host')

    parser.add_argument('--terraref_site', type=str,
            default=os.environ.get('TERRAREF_SITE', 'ua-mac'),
            help='site name (default=TERRAREF_SITE | ua-mac)')

    parser.add_argument('--terraref_sensor', type=str,
            default=os.environ.get('TERRAREF_SENSOR', ''),
            help='sensor name (default=None)')


def exact_p(pattern):

    return '^{}$'.format(pattern)


class Sensors():
    """Simple class used to save components of the file system path
    where sensor data will be stored.
    """

    def __init__(self, base, station, sensor='', stations=STATIONS):
        """Initialize basic path elements from env and cmdline."""

        self.stations = stations
        self.base = base.rstrip('/')

        if station in self.stations.keys():
            self.station = station
        else:
            raise AttributeError('unknown station name "{}"'.format(station))

        self._sensor = sensor


    @property
    def sensor(self):

        if self._sensor:
            return self._sensor
        else:
            raise RuntimeError('sensor not set')
        

    @sensor.setter
    def sensor(value):
        if value in self.stations[self.station].keys():
            self._sensor = value
        else:
            raise AttributeError('unknown sensor name "{}"'.format(value))


    def get_sensor_path(self, timestamp, sensor='', filename='',
                        opts=None, ext='', plot='', subsensor=''):
        """Get the appropritate path for writing sensor data

        Args:
          timestamp (datetime str): timestamp string
          sensor (str): overrides instance sensor variable, optional
          filename (str): overrides default filename, optional
          opts (list): list of filename extensions, optional
          ext (str): overrides the default filename extension, optional

        Returns:
          (str) full path to file

        Notes:
          When no filename is given, get_sensor_path returns the desired
          path with pre-defined, well-known filename. If filename is not
          given and no pre-defined filename exists, then a RuntimeError
          is raised.

          If a filename is supplied, it must match a pre-defined pattern.
          A RuntimeError is raised if the filename does not match the
          pattern.
        """

        # override class sensor
        if not sensor:
            sensor = self.sensor
        # split timestamp into date and hour-minute-second components
        if timestamp.find('__') > -1:
            date, hms = timestamp.split('__')
        else:
            date = timestamp
            hms = ''

        # Get regex patterns for this site/sensor
        try:
            s = self.stations[self.station][sensor]
        except KeyError:
            raise RuntimeError('sensor {} does not exist'.format(sensor))

        # create opt string
        if opts:
            opts = '_'.join(['']+opts)
        else:
            opts = ''

        # pattern should be completed with regex string using format
        if filename:
            if 'pattern' in s:
                pattern = exact_p(s['pattern']).format(sensor='\D*',
                        station='\D*', date=date_p, time=full_time_p,
                        timestamp=full_date_p, opts='\D*')

                result = re.match(pattern, filename)
                if result == None:
                    raise RuntimeError('The filename given does not match the correct pattern')

        else:
            filename = s['pattern'].format(station=self.station,
                    sensor=sensor, timestamp=timestamp, date=date, time=hms,
                    opts=opts)

        # replace default extension in filename
        if ext:
            filename = '.'.join([os.path.splitext(filename)[0], 
                                ext.lstrip('.')])

        path = s['template'].format(base=self.base, station=self.station,
                                    sensor=sensor, timestamp=timestamp,
                                    date=date, time=hms, filename=filename, plot=plot, subsensor=subsensor)
        return path


    def get_sensor_path_by_dataset(self, dsname, sensor='', ext='', 
                                   opts=None, hms=''):
        """Get the appropritate path for writing sensor data

        Args:
          datasetname (sensor - date__timestamp str):
              e.g. VNIR - 2017-06-28__23-48-28-435
          sensor (str): sensor name, may be a product name for Level_1
          opts (list of strs): any suffixes to apply to generated filename
          hms (str): allows one to add/override a timestamp
              if not included in datasetname (e.g. EnvironmentLogger)

        Returns:
          (str) full path
        """

        # Split dataset into sensorname and timestamp portions
        if dsname.find(" - ") > -1:
            sensorname, time = dsname.split(" - ")
        else:
            sensorname = dsname
            time = "2017-01-01"

        # Override/add timestamp if necessary
        if hms:
            if time.find("__") > -1:
                date = time.split("__")[0]
            else:
                date = time
            time = date + "__" + hms

        # Override dataset sensor name with provided name if given
        # (e.g. use timestamp of raw while getting Level_1 output)
        if not sensor:
            sensor = sensorname

        return self.get_sensor_path(time, sensor=sensor, opts=opts, ext=ext)


    def get_fixed_jsonpath_for_sensor(self, site=None, sensor=None):
        """Return the Clowder dataset ID for the fixed sensor information"""

        if not site:
            site = self.station
        if not sensor:
            sensor = self.sensor

        return self.stations[site][sensor]['fixed_metadata_json']


    def create_sensor_path(self, timestamp, sensor='', filename='',
                        opts=None, ext='', plot='', subsensor=''):
        """Return the full path for the sensor data
        
        Note: this function is similar to get_sensor_path and takes
        all the same arguments but has a side-effect of creating
        any missing directories in the path.
        """

        path = self.get_sensor_path(timestamp, sensor, filename,
                                        opts, ext, plot, subsensor)
        dirs = os.path.dirname(path) 
        if not os.path.exists(dirs):
            try:
                os.makedirs(dirs)
            except:
                # If another extractor created the dir structure in meantime, we can continue
                pass

        return path


    def get_sites(self):
        """Get all sites (stations) listed."""

        return self.stations.keys()


    def get_experiment(self, date):
        """
        Return the experiment metadata associated with the specified date.
        """
        experiments = get_experiments()

        # We only care about date portion if timestamp is given
        if date.find("__") > -1:
            date = date.split("__")[0]

        ds_time = datetime.datetime.strptime(date, "%Y-%m-%d")
        matched_exps = []
        for exp in experiments:
            begin = datetime.datetime.strptime(exp['start_date'], "%Y-%m-%d")
            end = datetime.datetime.strptime(exp['end_date'], "%Y-%m-%d")

            if ds_time >= begin and ds_time <= end:
                matched_exps.append(exp)
        return matched_exps


    def get_season(self, date):
        exps = self.get_experiment(date)
        if len(exps) > 0:
            return exps[0]['name'][:exps[0]['name'].find(':')]
        else:
            return None


    def get_sensors(self, station=''):
        """Get all sensors for a given station."""

        # TODO: allow the KeyError on bad station name?
        if station:
            return self.stations[station].keys()
        else:
            return self.stations[self.station].keys()


    def get_display_name(self, sensor=''):
        """Get display name for a sensor."""

        if sensor:
            if 'display' in self.stations[self.station][sensor]:
                return self.stations[self.station][sensor]['display']
            else:
                return sensor
        else:
            if 'display' in self.stations[self.station][self.sensor]:
                return self.stations[self.station][self.sensor]['display']
            else:
                return self.sensor
